//==================================================================
//
//	File name	:	panel.js
//
//	Summary		:	Panel, GroupBox, Tab classes
//
//	Ceated		:	2018/11/28
//	Written by	:	Yuji Iihara
//	Copyright	:	Softinc.
//
//==================================================================
"use strict";

(function (panel) {
	//==========================================================
	//	Global
	//==========================================================
	// Const
	panel.STYLE_DEFAULT = 0;
	panel.STYLE_INSET = 1;
	panel.STYLE_OUTSET = 2;
	panel.OVERFLOW_DEFAULT = 0;
	panel.OVERFLOW_HIDDEN = 1;
	panel.OVERFLOW_SCROLL = 2;
	panel.TYPE_H_SPLIT = 0;
	panel.TYPE_V_SPLIT = 1;
	panel.STYLE_FIXED = 0;
	panel.STYLE_VARIABLE = 1;
	panel.STYLE_ALIGN_RIGHT = 2;
	
	// Private variables
	let oResizingPanel = null;
	let iResizingPos = 0;
	
	//==========================================================
	//
	//	Class		:	Panel
	//
	//	Summary		:	Class of panel
	//
	//==========================================================
	panel.Panel = class extends jaswin.Container
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[String]
					oRect,			// * Size and position[ExpandableRect]
					iStyle,			// * Panel style[int]
					bEnabled,		// * Enabled[boolean]
					iOverflow,		// * Overflow[int]
					bPopup,			// Popup or not[boolean]
					bDroppable,		// Droppable or not[boolean]
					iDropType,		// Drop type[int]
					fpDrop)			// Function on dropped[function]
		{
			// Base class
			super(sID, oRect, bEnabled, -1);
			
			// Check parameters
			if(typeof iStyle != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID),
											"panel.js", "jaswin.panel.Panel.constructor");
			
			if((iStyle != panel.STYLE_DEFAULT)
			&& (iStyle != panel.STYLE_INSET)
			&& (iStyle != panel.STYLE_OUTSET))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, sID),
											"panel.js", "jaswin.panel.Panel.constructor");
			
			if(typeof iOverflow != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_OVERFLOW, sID),
											"panel.js", "jaswin.panel.Panel.constructor");
			
			if((iOverflow != panel.OVERFLOW_DEFAULT)
			&& (iOverflow != panel.OVERFLOW_HIDDEN)
			&& (iOverflow != panel.OVERFLOW_SCROLL))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_OVERFLOW, sID),
											"panel.js", "jaswin.panel.Panel.constructor");
			
			if(bPopup)
			{
				if(typeof bPopup != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_POPUP, sID),
												"panel.js", "jaswin.panel.Panel.constructor");
			}
			
			if(bDroppable)
			{
				if(typeof bDroppable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DROPPABLE, sID),
												"panel.js", "jaswin.panel.Panel.constructor");
				
				if(typeof iDropType != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DROP_TYPE, sID),
												"panel.js", "jaswin.panel.Panel.constructor");
				if((iDropType != jaswin.DROP_TEXT)
				&& (iDropType != jaswin.DROP_URL)
				&& (iDropType != jaswin.DROP_HTML)
				&& (iDropType != jaswin.DROP_FILE))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_DROP_TYPE, sID),
												"panel.js", "jaswin.panel.Panel.constructor");
				
				if(fpDrop)
				{
					if(typeof fpDrop != "function")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DROP, sID),
													"panel.js", "jaswin.panel.Panel.constructor");
				}
			}
			
			// Set properties
			this.style = iStyle;		// Panel style[int]
			this.popup = false;			// Popup or not[boolean]
			this.droppable = false;		// Droppable or not[boolean]
			this.dropType = iDropType;	// Drop type[int]
			this.onDrop = null;			// Function on drop[function(Argument:Panel class, e)]
			
			if(bPopup)
				this.popup = bPopup;
			
			if(bDroppable)
			{
				this.droppable = bDroppable;
				this.onDrop = fpDrop;
			}
			
			// Create element
			this.element = document.createElement("div");
			if(iStyle == panel.STYLE_DEFAULT)
				this.element.className = "panel";
			else if(iStyle == panel.STYLE_INSET)
				this.element.className = "panel-inset";
			else if(iStyle == panel.STYLE_OUTSET)
				this.element.className = "panel-outset";
			
			if(this.popup)
				this.element.style.boxShadow = "2px 2px 2px #a0a0a0";
			
			// Horizontal rect
			if(this.rect.anchorX == jaswin.ANCHOR_BOTH)
			{
				this.element.style.left = this.rect.left + "px";
				this.element.style.right = this.rect.right + "px";
			}
			else
			{
				if(this.rect.anchorX == jaswin.ANCHOR_LEFT)
					this.element.style.left = this.rect.left + "px";
				else if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
					this.element.style.right = this.rect.right + "px";
				
				if(this.style == panel.STYLE_DEFAULT)
					this.element.style.width = this.rect.width + "px";
				else
					this.element.style.width = this.rect.width - 2 + "px";
			}
			
			// Vertical rect
			if(this.rect.anchorY == jaswin.ANCHOR_BOTH)
			{
				this.element.style.top = this.rect.top + "px";
				this.element.style.bottom = this.rect.bottom + "px";
			}
			else
			{
				if(this.rect.anchorY == jaswin.ANCHOR_TOP)
					this.element.style.top = this.rect.top + "px";
				else if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
					this.element.style.bottom = this.rect.bottom + "px";
				
				if(this.style == panel.STYLE_DEFAULT)
					this.element.style.height = this.rect.height + "px";
				else
					this.element.style.height = this.rect.height - 2 + "px";
			}
			
			// Overflow
			if(iOverflow == panel.OVERFLOW_HIDDEN)
				this.element.style.overflow = "hidden";
			else if(iOverflow == panel.OVERFLOW_SCROLL)
				this.element.style.overflow = "auto";
			
			// Container of Panel is itself;
			this.container = this.element;
			
			// Event binding
			if(this.enabled && this.droppable)
			{
				this.element.addEventListener("dragover", this.dragOverEvent = (e) => {
					let iCounter;
					let bFound = false;
					
					for(iCounter = 0; iCounter < e.dataTransfer.types.length; iCounter++)
					{
						if((this.dropType == jaswin.DROP_TEXT)
						&& (e.dataTransfer.types[iCounter] == "text/plain"))
							bFound = true;
						else if((this.dropType == jaswin.DROP_URL)
						&& (e.dataTransfer.types[iCounter] == "text/uri-list"))
							bFound = true;
						else if((this.dropType == jaswin.DROP_HTML)
						&& (e.dataTransfer.types[iCounter] == "text/html"))
							bFound = true;
						else if((this.dropType == jaswin.DROP_FILE)
						&& (e.dataTransfer.types[iCounter] == "Files"))
							bFound = true;
					}
					if(!bFound)
						return;
					
					e.dataTransfer.dropEffect = "copy";
					e.stopPropagation();
					e.preventDefault();
				}, false);
				this.element.addEventListener("drop", this.dropEvent = (e) => {
					e.stopPropagation();
					e.preventDefault();
					if(this.onDrop)
						this.onDrop(this, e);
				}, false);
			}
		}
		
		//==================================================
		//	Set parent class
		//==================================================
		setParent(oParent)
		{
			super.setParent(oParent);
			
			let oParentWnd = jaswin.findParentWindow(this);
			if((jaswin.frame)
			&& (oParentWnd != global))
			{
				if(this.popup)
					this.element.style.zIndex = 50 + jaswin.frame.windowCount + oParentWnd.modalLevel;
			}
		}
		
		//==================================================
		//	Enable this panel
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled && this.droppable)
			{
				if(!this.dragOverEvent)
				{
					this.element.addEventListener("dragover", this.dragOverEvent = (e) => {
						let iCounter;
						let bFound = false;
						
						for(iCounter = 0; iCounter < e.dataTransfer.types.length; iCounter++)
						{
							if((this.dropType == jaswin.DROP_TEXT)
							&& (e.dataTransfer.types[iCounter] == "text/plain"))
								bFound = true;
							else if((this.dropType == jaswin.DROP_URL)
							&& (e.dataTransfer.types[iCounter] == "text/uri-list"))
								bFound = true;
							else if((this.dropType == jaswin.DROP_HTML)
							&& (e.dataTransfer.types[iCounter] == "text/html"))
								bFound = true;
							else if((this.dropType == jaswin.DROP_FILE)
							&& (e.dataTransfer.types[iCounter] == "Files"))
								bFound = true;
						}
						if(!bFound)
							return;
						
						e.dataTransfer.dropEffect = "copy";
						e.stopPropagation();
						e.preventDefault();
					}, false);
				}
				if(!this.dropEvent)
				{
					this.element.addEventListener("drop", this.dropEvent = (e) => {
						e.stopPropagation();
						e.preventDefault();
						if(this.onDrop)
							this.onDrop(this, e);
					}, false);
				}
			}
		}
		
		//==================================================
		//	Disable this panel
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			if(this.dragOverEvent)
			{
				this.element.removeEventListener("dragover", this.dragOverEvent, false);
				this.dragOverEvent = null;
			}
			if(this.dropEvent)
			{
				this.element.removeEventListener("drop", this.dropEvent, false);
				this.dropEvent = null;
			}
		}
		
		//==================================================
		//	Set background color
		//==================================================
		setBackgroundColor(sColor)
		{
			if(typeof sColor != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_BACKGROUND, this.id),
											"panel.js", "jaswin.panel.Panel.setBackgroundColor");
			
			this.container.style.backgroundColor = sColor;
		}
		
		//==================================================
		//	Set size
		//==================================================
		setSize(iWidth, iHeight)
		{
			if(typeof iWidth != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_WIDTH, this.id),
											"panel.js", "jaswin.panel.Panel.setSize");
			if(typeof iHeight != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_HEIGHT, this.id),
											"panel.js", "jaswin.panel.Panel.setSize");
			
			if(this.rect.anchorX != jaswin.ANCHOR_BOTH)
			{
				this.rect.width = iWidth;
				if(this.style == panel.STYLE_DEFAULT)
					this.element.style.width = this.rect.width + "px";
				else
					this.element.style.width = this.rect.width - 2 + "px";
			}
			if(this.rect.anchorY != jaswin.ANCHOR_BOTH)
			{
				this.rect.height = iHeight;
				if(this.style == panel.STYLE_DEFAULT)
					this.element.style.height = this.rect.height + "px";
				else
					this.element.style.height = this.rect.height - 2 + "px";
			}
		}
		
		//==================================================
		//	Notify destroy window
		//==================================================
		notifyDestroyWindow(oWindow)
		{
			if(!jaswin.frame)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_INCLUDE, "frame.js"),
											"panel.js", "jaswin.panel.Panel.notifyDestroyWindow");
			
			if((typeof oWindow != "object")
			|| (!oWindow instanceof jaswin.frame.Window))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_WINDOW, this.id),
											"panel.js", "jaswin.panel.Panel.notifyDestroyWindow");
			
			let iCounter;
			
			for(iCounter = 0; iCounter < this.controls.length; iCounter++)
			{
				if(this.controls[iCounter] == oWindow)
				{
					this.controls.splice(iCounter, 1);
					return;
				}
			}
		}
		
		//==================================================
		//	Set context menu to this panel
		//==================================================
		setContextMenu(oMenu, fpContext)
		{
			super.setContextMenu(oMenu, fpContext);
			
			let oParentWnd = jaswin.findParentWindow(this);
			oParentWnd.container.appendChild(oMenu.element);
			
			oMenu.setParent(this, 0);
			
			if(!this.contextMenuEvent)
			{
				this.container.addEventListener("contextmenu", this.contextMenuEvent = (e) => {
					this.showContextMenu(e);
				}, false);
			}
		}
		
		//==================================================
		//	Remove context menu from this panel
		//==================================================
		removeContextMenu()
		{
			super.removeContextMenu();
			
			if(this.contextMenuEvent)
				this.container.removeEventListener("contextmenu", this.contextMenuEvent, false);
			this.contextMenuEvent = null;
		}
		
		//==================================================
		//	Show context menu
		//==================================================
		showContextMenu(e)
		{
			let oElement = e.target;
			
			if(oElement != this.container)
				return;
			
			if(this.onContextMenu)
			{
				if(!this.onContextMenu(this))
					return;
			}
			
			this.contextMenu.showContext(e);
		}
		
		//==================================================
		//	Set droppable
		//==================================================
		setDroppable(fpDrop)
		{
			if(fpDrop)
			{
				if(typeof fpDrop != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DROP, this.id),
												"panel.js", "jaswin.panel.Panel.setDroppable");
			}
			
			this.droppable = true;
			this.onDrop = fpDrop;
			
			if((this.enabled) && (this.parentEnabled))
			{
				if(!this.dragOverEvent)
				{
					this.element.addEventListener("dragover", this.dragOverEvent = (e) => {
						e.preventDefault();
					}, false);
				}
				if(this.dropEvent)
					this.element.removeEventListener("drop", this.dropEvent, false);
				
				this.element.addEventListener("drop", this.dropEvent = (e) => {
					if(this.onDrop)
						this.onDrop(this, e);
					e.preventDefault();
				}, false);
			}
		}
		
		//==================================================
		//	Set undroppable
		//==================================================
		setUnDroppable()
		{
			this.droppable = false;
			this.onDrop = null;
			
			if(this.dragOverEvent)
			{
				this.element.removeEventListener("dragover", this.dragOverEvent, false);
				this.dragOverEvent = null;
			}
			if(this.dropEvent)
			{
				this.element.removeEventListener("drop", this.dropEvent, false);
				this.dropEvent = null;
			}
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	SplitPanel
	//
	//	Summary		:	Class of SplitPanel
	//
	//==========================================================
	panel.SplitPanel = class extends jaswin.Container
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[String]
					oRect,			// * Size and position[ExpandableRect]
					iType,			// * Type of separation[int]
					bEnabled,		// * Enabled[boolean]
					iDelimiterPos,	// * Position of delimiter[int]
					bResizable,		// * Resizable or not[boolean]
					iaStyles,		// * Panel style[Array of int]
					iaOverflows,	// * Overflow[int]
					iMinPos,		// Minimum position of delimiter[int]
					iMaxPos)		// Maximum position of delimiter[int]
		{
			// Base class
			super(sID, oRect, bEnabled, -1);
			
			// Check parameters
			if(typeof iType != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_SPLIT, sID),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			if((iType != panel.TYPE_H_SPLIT)
			&& (iType != panel.TYPE_V_SPLIT))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_SPLIT, sID),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			
			if(typeof iDelimiterPos != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_DELIMITER, sID),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			
			if(typeof bResizable != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_RESIZABLE, sID),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			
			if(!Array.isArray(iaStyles))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			if(iaStyles.length < 2)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_NUMBER_STYLE, sID),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			if(typeof iaStyles[0] != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID + "_0"),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			if(typeof iaStyles[1] != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID + "_1"),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			
			if(!Array.isArray(iaOverflows))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_OVERFLOW, sID),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			if(iaOverflows.length < 2)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_NUMBER_OVERFLOW, sID),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			if(typeof iaOverflows[0] != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_OVERFLOW, sID + "_0"),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			if(typeof iaOverflows[1] != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_OVERFLOW, sID + "_1"),
											"panel.js", "jaswin.panel.SplitPanel.constructor");
			
			if(iMinPos)
			{
				if(typeof iMinPos != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_MIN_POSITION, sID),
												"panel.js", "jaswin.panel.SplitPanel.constructor");
			}
			
			if(iMaxPos)
			{
				if(typeof iMaxPos != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_MAX_POSITION, sID),
												"panel.js", "jaswin.panel.SplitPanel.constructor");
			}
			
			// Set properties
			this.type = iType;					// Separation type[int]
			this.delimiterPos = iDelimiterPos;	// Position of delimiter[int]
			this.minPos = -1;					// Minimum position of delimiter[int]
			this.maxPos = -1;					// Maximum position of delimiter[int]
			this.delimiter = null;				// Delemiter element[DOM Element]
			this.resizable = bResizable;		// Resizable or not[boolean]
			
			if(iMinPos) this.minPos = iMinPos;
			if(iMaxPos) this.maxPos = iMaxPos;
			
			// Create element
			this.element = document.createElement("div");
			this.element.className = "panel";
			
			// Horizontal rect
			if(this.rect.anchorX == jaswin.ANCHOR_BOTH)
			{
				this.element.style.left = this.rect.left + "px";
				this.element.style.right = this.rect.right + "px";
			}
			else
			{
				if(this.rect.anchorX == jaswin.ANCHOR_LEFT)
					this.element.style.left = this.rect.left + "px";
				else if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
					this.element.style.right = this.rect.right + "px";
				this.element.style.width = this.rect.width + "px";
			}
			
			// Vertical rect
			if(this.rect.anchorY == jaswin.ANCHOR_BOTH)
			{
				this.element.style.top = this.rect.top + "px";
				this.element.style.bottom = this.rect.bottom + "px";
			}
			else
			{
				if(this.rect.anchorY == jaswin.ANCHOR_TOP)
					this.element.style.top = this.rect.top + "px";
				else if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
					this.element.style.bottom = this.rect.bottom + "px";
				this.element.style.height = this.rect.height + "px";
			}
			
			// Container of Panel is itself;
			this.container = this.element;
			
			// Make first panel
			let oExpRect;
			let oPanel;
			
			if(this.type == panel.TYPE_H_SPLIT)
			{
				oExpRect = new jaswin.ExpandableRect(jaswin.ANCHOR_LEFT, jaswin.ANCHOR_BOTH,
													0, 0,
													0, 0,
													this.delimiterPos, 0);
			}
			else if(this.type == panel.TYPE_V_SPLIT)
			{
				oExpRect = new jaswin.ExpandableRect(jaswin.ANCHOR_BOTH, jaswin.ANCHOR_TOP,
													0, 0,
													0, 0,
													0, this.delimiterPos);
			}
			oPanel = new panel.Panel(this.id + "_0", oExpRect, iaStyles[0], true, iaOverflows[0]);
			this.container.appendChild(oPanel.element);
			this.controls[0] = oPanel;
			
			// Make delimiter
			this.delimiter = document.createElement("div");
			if(this.type == panel.TYPE_H_SPLIT)
			{
				this.delimiter.className = "panel-h-delimiter";
				this.delimiter.style.left = this.delimiterPos + "px";
			}
			else if(this.type == panel.TYPE_V_SPLIT)
			{
				this.delimiter.className = "panel-v-delimiter";
				this.delimiter.style.top = this.delimiterPos + "px";
			}
			this.container.appendChild(this.delimiter);
			
			if(this.enabled && this.resizable)
			{
				this.delimiter.addEventListener("mousedown", this.mouseDownEvent = (e) => {
					startResizePanel(this, e);
				} , false);
				
				if(this.type == panel.TYPE_H_SPLIT)
					this.delimiter.style.cursor = "col-resize";
				else
					this.delimiter.style.cursor = "row-resize";
			}
			
			// Make second panel
			if(this.type == panel.TYPE_H_SPLIT)
			{
				oExpRect = new jaswin.ExpandableRect(jaswin.ANCHOR_BOTH, jaswin.ANCHOR_BOTH,
													this.delimiterPos + 2, 0,
													0, 0,
													0, 0);
			}
			else if(this.type == panel.TYPE_V_SPLIT)
			{
				oExpRect = new jaswin.ExpandableRect(jaswin.ANCHOR_BOTH, jaswin.ANCHOR_BOTH,
													0, 0,
													this.delimiterPos + 2, 0,
													0, 0);
			}
			oPanel = new panel.Panel(this.id + "_1", oExpRect, iaStyles[1], true, iaOverflows[1]);
			this.container.appendChild(oPanel.element);
			this.controls[1] = oPanel;
		}
		
		//==================================================
		//	Set parent class
		//==================================================
		setParent(oParent)
		{
			super.setParent(oParent);
			
			this.controls[0].setParent(this);
			this.controls[1].setParent(this);
		}
		
		//==================================================
		//	Enable this panel
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled && this.resizable)
			{
				if(!this.mouseDownEvent)
				{
					this.delimiter.addEventListener("mousedown", this.mouseDownEvent = (e) => {
						startResizePanel(this, e);
					} , false);
				}
				
				if(this.type == panel.TYPE_H_SPLIT)
					this.delimiter.style.cursor = "col-resize";
				else
					this.delimiter.style.cursor = "row-resize";
			}
		}
		
		//==================================================
		//	Disable this panel
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			if(this.resizable && this.mouseDownEvent)
			{
				this.delimiter.removeEventListener("mousedown", this.mouseDownEvent, false);
				this.mouseDownEvent = null;
			}
			this.delimiter.style.cursor = "default";
		}
		
		//==================================================
		//	Add control to this container
		//==================================================
		addControl(oControl, oParentElement)
		{
			return;
		}
		
		//==================================================
		//	Remove control from this container
		//==================================================
		removeControl(oControl)
		{
			return;
		}
		
		//==================================================
		//	Resizable or not
		//==================================================
		setResizable(bResizable)
		{
			if(typeof bResizable != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"panel.js", "jaswin.panel.SplitPanel.setResizable");
			if(this.resizable == bResizable)
				return;
			
			this.resizable = bResizable;
			if((this.resizable) && (this.enabled))
			{
				if(!this.mouseDownEvent)
				{
					this.delimiter.addEventListener("mousedown", this.mouseDownEvent = (e) => {
						startResizePanel(this, e);
					} , false);
				}
				
				if(this.type == panel.TYPE_H_SPLIT)
					this.delimiter.style.cursor = "col-resize";
				else
					this.delimiter.style.cursor = "row-resize";
			}
			else if(!this.resizable && this.mouseDownEvent)
			{
				this.delimiter.removeEventListener("mousedown", this.mouseDownEvent, false);
				this.mouseDownEvent = null;
			}
		}
		
		//==================================================
		//	Get child panel
		//==================================================
		getChildPanel(iIndex)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"panel.js", "jaswin.panel.SplitPanel.getChildPanel");
			if((iIndex != 0) && (iIndex != 1))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, this.id),
											"panel.js", "jaswin.panel.SplitPanel.getChildPanel");
			
			return this.controls[iIndex];
		}
		
		//==================================================
		//	Resize panels
		//==================================================
		resizePanel(iGap)
		{
			if(this.delimiterPos + iGap < 0)
			{
				// Under 0
				this.element.removeEventListener("mouseup", endResizePanel, false);
				this.element.removeEventListener("mousemove", resizePanel, false);
				if(this.type == panel.TYPE_H_SPLIT)
					this.delimiter.className = "panel-h-delimiter";
				else if(this.type == panel.TYPE_V_SPLIT)
					this.delimiter.className = "panel-v-delimiter";
				oResizingPanel = null;
				return;
			}
			
			if(this.type == panel.TYPE_H_SPLIT)
			{
				if(this.delimiterPos + iGap + 2 > this.element.scrollWidth)
				{
					// Over size
					this.element.removeEventListener("mouseup", endResizePanel, false);
					this.element.removeEventListener("mousemove", resizePanel, false);
					this.delimiter.className = "panel-h-delimiter";
					oResizingPanel = null;
					return;
				}
			}
			else if(this.type == panel.TYPE_V_SPLIT)
			{
				if(this.delimiterPos + iGap + 2 > this.element.scrollHeight)
				{
					// Maximum
					this.element.removeEventListener("mouseup", endResizePanel, false);
					this.element.removeEventListener("mousemove", resizePanel, false);
					this.delimiter.className = "panel-v-delimiter";
					oResizingPanel = null;
					return;
				}
			}
			
			this.delimiterPos += iGap;
			
			if(this.type == panel.TYPE_H_SPLIT)
			{
				this.controls[0].rect.width = this.delimiterPos - 2;
				if(this.controls[0].style == panel.STYLE_DEFAULT)
					this.controls[0].element.style.width = this.delimiterPos + "px";
				else
					this.controls[0].element.style.width = this.delimiterPos - 2 + "px";
				this.delimiter.style.left = this.delimiterPos + "px";
				this.controls[1].rect.left = this.delimiterPos + 2;
				this.controls[1].element.style.left = (this.delimiterPos + 2) + "px";
			}
			else if(this.type == panel.TYPE_V_SPLIT)
			{
				this.controls[0].rect.height = this.delimiterPos - 2;
				if(this.controls[0].style == panel.STYLE_DEFAULT)
					this.controls[0].element.style.height = this.delimiterPos + "px";
				else
					this.controls[0].element.style.height = this.delimiterPos - 2 + "px";
				this.delimiter.style.top = this.delimiterPos + "px";
				this.controls[1].rect.top = this.delimiterPos + 2;
				this.controls[1].element.style.top = (this.delimiterPos + 2) + "px";
			}
		}
		
		//==================================================
		//	Set context menu to this panel
		//==================================================
		setContextMenu(oMenu, fpContext)
		{
			return;
		}
		
		//==================================================
		//	Remove context menu from this container
		//==================================================
		removeContextMenu()
		{
			return;
		}
	}
	
	
	//==========================================================
	//
	//	Function	:	startResizePanel
	//
	//	Summary		:	Event handler of start resizing panel
	//
	//==========================================================
	function startResizePanel(oPanel, e)
	{
		oResizingPanel = oPanel;
		
		if(oPanel.type == panel.TYPE_H_SPLIT)
		{
			iResizingPos = e.clientX;
			oPanel.delimiter.className = "panel-h-delimiter-on";
		}
		else if(oPanel.type == panel.TYPE_V_SPLIT)
		{
			iResizingPos = e.clientY;
			oPanel.delimiter.className = "panel-v-delimiter-on";
		}
		
		oPanel.element.addEventListener("mouseup", endResizePanel, false);
		oPanel.element.addEventListener("mousemove", resizePanel, false);
	}
	
	
	//==========================================================
	//
	//	Function	:	resizePanel
	//
	//	Summary		:	Event handler of resizing panel
	//
	//==========================================================
	function resizePanel(e)
	{
		let oElement = e.target;
		let iCurPos;
		let oSelection = window.getSelection();
		
		oSelection.removeAllRanges();
		
		if(!oResizingPanel)
			return;
		if(!oResizingPanel.element.contains(oElement))
		{
			oResizingPanel.element.removeEventListener("mouseup", endResizePanel, false);
			oResizingPanel.element.removeEventListener("mousemove", resizePanel, false);
			
			if(oResizingPanel.type == panel.TYPE_H_SPLIT)
				oResizingPanel.delimiter.className = "panel-h-delimiter";
			else if(oResizingPanel.type == panel.TYPE_V_SPLIT)
				oResizingPanel.delimiter.className = "panel-v-delimiter";
			
			oResizingPanel = null;
			return;
		}
		
		if(oResizingPanel.type == panel.TYPE_H_SPLIT)
			iCurPos = e.clientX;
		else if(oResizingPanel.type == panel.TYPE_V_SPLIT)
			iCurPos = e.clientY;
		
		if((oResizingPanel.minPos >= 0) && (iCurPos - iResizingPos < 0))
		{
			if(oResizingPanel.delimiterPos + iCurPos - iResizingPos < oResizingPanel.minPos)
			{
				// Under minimum position
				oResizingPanel.element.removeEventListener("mouseup", endResizePanel, false);
				oResizingPanel.element.removeEventListener("mousemove", resizePanel, false);
				
				if(oResizingPanel.type == panel.TYPE_H_SPLIT)
					oResizingPanel.delimiter.className = "panel-h-delimiter";
				else if(oResizingPanel.type == panel.TYPE_V_SPLIT)
					oResizingPanel.delimiter.className = "panel-v-delimiter";
				
				oResizingPanel = null;
				return;
			}
		}
		else if((oResizingPanel.maxPos >= 0) && (iCurPos - iResizingPos > 0))
		{
			if(oResizingPanel.delimiterPos + iCurPos - iResizingPos > oResizingPanel.maxPos)
			{
				// Over maximum position
				oResizingPanel.element.removeEventListener("mouseup", endResizePanel, false);
				oResizingPanel.element.removeEventListener("mousemove", resizePanel, false);
				
				if(oResizingPanel.type == panel.TYPE_H_SPLIT)
					oResizingPanel.delimiter.className = "panel-h-delimiter";
				else if(oResizingPanel.type == panel.TYPE_V_SPLIT)
					oResizingPanel.delimiter.className = "panel-v-delimiter";
				
				oResizingPanel = null;
				return;
			}
		}
		
		oResizingPanel.resizePanel(iCurPos - iResizingPos);
		iResizingPos = iCurPos;
	}
	
	
	//==========================================================
	//
	//	Function	:	endResizePanel
	//
	//	Summary		:	Event handler of end resizing panel
	//
	//==========================================================
	function endResizePanel(e)
	{
		if(!oResizingPanel)
			return;
		
		oResizingPanel.element.removeEventListener("mouseup", endResizePanel, false);
		oResizingPanel.element.removeEventListener("mousemove", resizePanel, false);
		
		if(oResizingPanel.type == panel.TYPE_H_SPLIT)
			oResizingPanel.delimiter.className = "panel-h-delimiter";
		else if(oResizingPanel.type == panel.TYPE_V_SPLIT)
			oResizingPanel.delimiter.className = "panel-v-delimiter";
		
		oResizingPanel = null;
	}
	
	
	//==========================================================
	//
	//	Class		:	GroupBox
	//
	//	Summary		:	Class of GroupBox
	//
	//==========================================================
	panel.GroupBox = class extends jaswin.Container
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[String]
					oRect,			// * Size and position[ExpandableRect]
					sTitle,			// * Title[String]
					bEnabled)		// * Enabled[boolean]
		{
			// Base class
			super(sID, oRect, bEnabled, -1);
			
			// Check parameters
			if(typeof sTitle != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TITLE, sID),
											"panel.js", "jaswin.panel.GroupBox.constructor");
			
			// Set properties
			this.label = null;				// Title label[DOM Element]
			
			// Create element
			this.element = document.createElement("div");
			this.element.className = "groupbox";
			
			// Horizontal rect
			if(this.rect.anchorX == jaswin.ANCHOR_BOTH)
			{
				this.element.style.left = this.rect.left + "px";
				this.element.style.right = this.rect.right + "px";
			}
			else
			{
				if(this.rect.anchorX == jaswin.ANCHOR_LEFT)
					this.element.style.left = this.rect.left + "px";
				else if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
					this.element.style.right = this.rect.right + "px";
				this.element.style.width = this.rect.width + "px";
			}
			
			// Vertical rect
			if(this.rect.anchorY == jaswin.ANCHOR_BOTH)
			{
				this.element.style.top = this.rect.top + "px";
				this.element.style.bottom = this.rect.bottom + "px";
			}
			else
			{
				if(this.rect.anchorY == jaswin.ANCHOR_TOP)
					this.element.style.top = this.rect.top + "px";
				else if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
					this.element.style.bottom = this.rect.bottom + "px";
				this.element.style.height = this.rect.height + "px";
			}
			
			let oLine = document.createElement("div");
			oLine.className = "groupbox-line";
			this.element.appendChild(oLine);
			
			this.label = document.createElement("span");
			this.label.className = "groupbox-title";
			this.label.innerHTML = sTitle;
			this.element.appendChild(this.label);
			
			this.container = document.createElement("div");
			this.container.className = "groupbox-container";
			this.element.appendChild(this.container);
		}
		
		//==================================================
		//	Enable this groupbox
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled)
				this.label.className = "groupbox-title";
		}
		
		//==================================================
		//	Disable this groupbox
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			this.label.className = "groupbox-title-disabled";
		}
		
		//==================================================
		//	Set title to this groupbox
		//==================================================
		setTitle(sTitle)
		{
			if(typeof sTitle != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TITLE, this.id),
											"panel.js", "jaswin.panel.GroupBox.setTitle");
			
			this.label.innerHTML = sTitle;
		}
		
		//==================================================
		//	Set context menu to this groupbox
		//==================================================
		setContextMenu(oMenu, fpContext)
		{
			super.setContextMenu(oMenu, fpContext);
			
			let oParentWnd = jaswin.findParentWindow(this);
			oParentWnd.container.appendChild(oMenu.element);
			
			oMenu.setParent(this, 0);
			
			if(!this.contextMenuEvent)
			{
				this.container.addEventListener("contextmenu", this.contextMenuEvent = (e) => {
					this.showContextMenu(e);
				}, false);
			}
		}
		
		//==================================================
		//	Remove context menu from this groupbox
		//==================================================
		removeContextMenu()
		{
			super.removeContextMenu();
			
			if(this.contextMenuEvent)
				this.container.removeEventListener("contextmenu", this.contextMenuEvent, false);
			this.contextMenuEvent = null;
		}
		
		//==================================================
		//	Show context menu
		//==================================================
		showContextMenu(e)
		{
			let oElement = e.target;
			
			if(oElement != this.container)
				return;
			
			if(this.onContextMenu)
			{
				if(!this.onContextMenu(this))
					return;
			}
			
			this.contextMenu.showContext(e);
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	Tab
	//
	//	Summary		:	Class of tab
	//
	//==========================================================
	panel.Tab = class extends jaswin.Container
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[String]
					oRect,			// * Size and position[ExpandableRect]
					iStyle,			// * Panel style[int]
					bEnabled,		// * Enabled[boolean]
					saLabels,		// Index labels[Array of string]
					iStartIndex,	// Start index of active tab[int]
					fpSelect)		// Function on select[function]
		{
			// Base class
			super(sID, oRect, bEnabled, -1);
			
			// Check parameters
			if(typeof iStyle != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID),
											"panel.js", "jaswin.panel.Tab.constructor");
			
			if((!iStyle & panel.STYLE_VARIABLE) && (saLabels == null))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, sID),
											"panel.js", "jaswin.panel.Tab.constructor");
			
			let iCnt = 0;
			let iCounter = 0;
			let iIndex = 0;
			
			if(saLabels)
			{
				if(!Array.isArray(saLabels))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, sID),
												"panel.js", "jaswin.panel.Tab.constructor");
				if((!iStyle & panel.STYLE_VARIABLE) && (saLabels.length == 0))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_NUMBER_LABEL, sID),
												"panel.js", "jaswin.panel.Tab.constructor");
				iCnt = saLabels.length;
				for(iCounter = 0; iCounter < iCnt; iCounter++)
				{
					if(typeof saLabels[iCounter] != "string")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, sID + "_1"),
													"panel.js", "jaswin.panel.Tab.constructor");
				}
			}
			
			if(iCnt)
			{
				if(iStartIndex)
				{
					if(typeof iStartIndex != "number")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, sID),
													"panel.js", "jaswin.panel.Tab.constructor");
					
					if((iStartIndex < 0) || (iStartIndex >= iCnt))
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, sID),
													"panel.js", "jaswin.panel.Tab.constructor");
					
					iIndex = iStartIndex;
				}
				else
					iIndex = 0;
			}
			else
				iIndex = 0;
			
			if(fpSelect)
			{
				if(typeof fpSelect != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CLICK, sID),
												"panel.js", "jaswin.panel.Tab.constructor");
			}
			
			// Set properties
			this.style = iStyle;				// Tab style[int]
			this.index = null;					// Element for indexes[DOM Element]
			this.indexContainer = null;			// Container for indexes[DOM Element]
			this.indexInner = null;				// Container inner for indexes[DOM Element]
			this.indexes = new Array();			// Indexes of tab[DOM Element]
			this.bgColors = new Array();		// Back ground color of tab items[String]
			this.selectedIndex = iIndex;		// Selected index[int]
			this.scrollLeftButton = null;		// Scroll left index button[boolean]
			this.scrollRightButton = null;		// Scroll right index button[boolean]
			this.adjustIndexHeight = false;		// Adjust index height or not[boolean]
			this.onSelect = fpSelect;			// Function on click[function(Argument:int)]
			
			// Create element
			this.element = document.createElement("div");
			this.element.style.position = "absolute";
			
			// Horizontal rect
			if(this.rect.anchorX == jaswin.ANCHOR_BOTH)
			{
				this.element.style.left = this.rect.left + "px";
				this.element.style.right = this.rect.right + "px";
			}
			else
			{
				if(this.rect.anchorX == jaswin.ANCHOR_LEFT)
					this.element.style.left = this.rect.left + "px";
				else if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
					this.element.style.right = this.rect.right + "px";
				this.element.style.width = this.rect.width + "px";
			}
			
			// Vertical rect
			if(this.rect.anchorY == jaswin.ANCHOR_BOTH)
			{
				this.element.style.top = this.rect.top + "px";
				this.element.style.bottom = this.rect.bottom + "px";
			}
			else
			{
				if(this.rect.anchorY == jaswin.ANCHOR_TOP)
					this.element.style.top = this.rect.top + "px";
				else if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
					this.element.style.bottom = this.rect.bottom + "px";
				this.element.style.height = this.rect.height + "px";
			}
			
			// Create contents
			let oPanel;
			let oExpRect;
			
			this.container = document.createElement("div");
			this.container.className = "tab-content";
			this.element.appendChild(this.container);
			
			for(iCounter = 0; iCounter < iCnt; iCounter++)
			{
				oExpRect = new jaswin.ExpandableRect(jaswin.ANCHOR_BOTH, jaswin.ANCHOR_BOTH,
													0, 0,
													0, 0,
													0, 0);
				oPanel = new panel.Panel(this.id + "_" + iCounter, oExpRect,
										panel.STYLE_DEFAULT, true,
										panel.OVERFLOW_DEFAULT);
				
				if(iCounter != this.selectedIndex)
					oPanel.hide();
				this.container.appendChild(oPanel.element);
				this.controls[iCounter] = oPanel;
				this.bgColors[iCounter] = "";
			}
			
			// Create indexes
			this.index = document.createElement("div");
			this.index.className = "tab-index";
			this.indexContainer = document.createElement("div");
			this.indexContainer.className = "tab-index-container";
			this.index.appendChild(this.indexContainer);
			this.indexInner = document.createElement("div");
			this.indexInner.className = "tab-index-inner";
			if(this.style & panel.STYLE_ALIGN_RIGHT)
				this.indexInner.style.right = "0px";
			else
				this.indexInner.style.left = "0px";
			
			this.indexContainer.appendChild(this.indexInner);
			
			for(iCounter = 0; iCounter < iCnt; iCounter++)
			{
				this.indexes[iCounter] = document.createElement("span");
				if(iCounter == this.selectedIndex)
				{
					if(this.enabled)
						this.indexes[iCounter].className = "tab-index-on";
					else
						this.indexes[iCounter].className = "tab-index-on-disabled";
				}
				else
				{
					if(this.enabled)
						this.indexes[iCounter].className = "tab-index-off";
					else
						this.indexes[iCounter].className = "tab-index-off-disabled";
				}
				this.indexes[iCounter].innerHTML = saLabels[iCounter];
				
				if(this.style & panel.STYLE_VARIABLE)
				{
					let oCloseSpan = document.createElement("span");
					if(this.enabled)
						oCloseSpan.className = "tab-index-close";
					else
						oCloseSpan.className = "tab-index-close-disabled";
					oCloseSpan.innerHTML = "x";
					this.indexes[iCounter].appendChild(oCloseSpan);
					
					if(this.enabled)
					{
						if(this.closeTabEvent)
							oCloseSpan.addEventListener("click", this.closeTabEvent, false);
						else
						{
							oCloseSpan.addEventListener("click", this.closeTabEvent = (e) => {
								this.closeTab(e);
							}, false);
						}
					}
				}
				
				if(this.enabled)
				{
					if(this.selectTabEvent)
						this.indexes[iCounter].addEventListener("click", this.selectTabEvent, false);
					else
					{
						this.indexes[iCounter].addEventListener("click", this.selectTabEvent = (e) => {
							this.select(e);
						}, false);
					}
				}
				
				this.indexInner.appendChild(this.indexes[iCounter]);
			}
			
			// Scroll button
			if(jaswin.button)
			{
				let oRect = new jaswin.FixedRect(jaswin.ANCHOR_LEFT, jaswin.ANCHOR_TOP,
												0, 4, 12, 18);
				this.scrollLeftButton = new jaswin.button.Button("Btn" + this.id + "_IndexLeft", oRect,
																jaswin.button.STYLE_DEFAULT | jaswin.button.STYLE_ICON,
																true,
																jaswin.ICON_TRIANGLE_LEFT, null,
																0, tabIndexScrollLeft);
				this.scrollLeftButton.hide();
				this.scrollLeftButton.content.childNodes[0].style.left = "-4px";
				this.index.appendChild(this.scrollLeftButton.element);
				
				oRect = new jaswin.FixedRect(jaswin.ANCHOR_RIGHT, jaswin.ANCHOR_TOP,
											0, 4, 12, 18);
				this.scrollRightButton = new jaswin.button.Button("Btn" + this.id + "_IndexRight", oRect,
																jaswin.button.STYLE_DEFAULT | jaswin.button.STYLE_ICON,
																true,
																jaswin.ICON_TRIANGLE_RIGHT, null,
																0, tabIndexScrollRight);
				this.scrollRightButton.hide();
				this.scrollRightButton.content.childNodes[0].style.left = "2px";
				this.index.appendChild(this.scrollRightButton.element);
				
				if(this.enabled)
				{
					this.index.addEventListener("mouseover", this.indexMouseMoveEvent = (e) => {
						this.hoverIndex();
					}, false);
					this.index.addEventListener("mouseout", this.indexMouseOutEvent = (e) => {
						this.outIndex(e);
					}, false);
				}
			}
			this.element.appendChild(this.index);
		}
		
		//==================================================
		//	Delete DOM element
		//==================================================
		destroy()
		{
			super.destroy();
			
			if(this.scrollLeftButton)
				this.scrollLeftButton.destroy();
			if(this.scrollRightButton)
				this.scrollRightButton.destroy();
		}
		
		//==================================================
		//	Set parent class
		//==================================================
		setParent(oParent)
		{
			super.setParent(oParent);
			
			// Adjust index padding
			let iCounter;
			let iHeight;
			let iPadding;
			for(iCounter = 0; iCounter < this.indexes.length; iCounter++)
			{
				if(iCounter == this.selectedIndex)
				{
					iHeight = this.indexes[iCounter].offsetHeight;
					iPadding = Math.floor((22 - iHeight) / 2);
					if(iPadding > 0)
					{
						this.indexes[iCounter].style.paddingTop = iPadding + 2 + "px";
						this.indexes[iCounter].style.paddingBottom = iPadding + "px";
						this.adjustIndexHeight = true;
					}
				}
				else
				{
					iHeight = this.indexes[iCounter].offsetHeight;
					iPadding = Math.floor((20 - iHeight) / 2);
					if(iPadding > 0)
					{
						this.indexes[iCounter].style.paddingTop = iPadding + "px";
						this.indexes[iCounter].style.paddingBottom = iPadding + "px";
						this.adjustIndexHeight = true;
					}
				}
				
				// Set parent
				this.controls[iCounter].setParent(this);
			}
			this.indexInner.style.top = "3px";
			
			// Adjust container position
			if(this.indexes.length == 0)
			{
				let oDummy = document.createElement("span");
				oDummy.className = "tab-index-on";
				oDummy.innerHTML = " ";
				this.indexInner.appendChild(oDummy);
				this.container.style.top = oDummy.offsetHeight - 1 + "px";
				this.indexInner.removeChild(oDummy);
			}
			else
				this.container.style.top = this.indexes[this.selectedIndex].offsetHeight - 1 + "px";
			
			// Set index scroll button
			if(this.scrollLeftButton)
				this.scrollLeftButton.setParent(this);
			if(this.scrollRightButton)
				this.scrollRightButton.setParent(this);
			
			let iInnerWidth = 0;
			
			for(iCounter = 0; iCounter < this.indexes.length; iCounter++)
				iInnerWidth += this.indexes[iCounter].offsetWidth;
			
			this.indexInner.style.width = (iInnerWidth + 4) + "px";
		}
		
		//==================================================
		//	Enable this tab
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled)
			{
				let iCounter;
				for(iCounter = 0; iCounter < this.controls.length; iCounter++)
				{
					if(this.selectTabEvent)
						this.indexes[iCounter].addEventListener("click", this.selectTabEvent, false);
					else
					{
						this.indexes[iCounter].addEventListener("click", this.selectTabEvent = (e) => {
							this.select(e);
						}, false);
					}
					this.controls[iCounter].enable(true);
					
					if(iCounter == this.selectedIndex)
						this.indexes[iCounter].className = "tab-index-on";
					else
						this.indexes[iCounter].className = "tab-index-off";
					
					if(this.style & panel.STYLE_VARIABLE)
					{
						let oCloseSpan = this.indexes[iCounter].getElementsByTagName("span")[0];
						
						oCloseSpan.className = "tab-index-close";
						if(this.closeTabEvent)
							oCloseSpan.addEventListener("click", this.closeTabEvent, false);
						else
						{
							oCloseSpan.addEventListener("click", this.closeTabEvent = (e) => {
								this.closeTab(e);
							}, false);
						}
					}
				}
				
				if(this.scrollLeftButton)
				{
					if(!this.indexMouseMoveEvent)
					{
						this.index.addEventListener("mouseover", this.indexMouseMoveEvent = (e) => {
							this.hoverIndex();
						}, false);
					}
					if(!this.indexMouseOutEvent)
					{
						this.index.addEventListener("mouseout", this.indexMouseOutEvent = (e) => {
							this.outIndex(e);
						}, false);
					}
				}
			}
		}
		
		//==================================================
		//	Disable this tab
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			let iCounter;
			
			for(iCounter = 0; iCounter < this.controls.length; iCounter++)
			{
				if(this.selectTabEvent)
					this.indexes[iCounter].removeEventListener("click", this.selectTabEvent, false);
				this.controls[iCounter].disable(true);
				
				if(iCounter == this.selectedIndex)
					this.indexes[iCounter].className = "tab-index-on-disabled";
				else
					this.indexes[iCounter].className = "tab-index-off-disabled";
				
				if(this.style & panel.STYLE_VARIABLE)
				{
					let oCloseSpan = this.indexes[iCounter].getElementsByTagName("span")[0];
					
					oCloseSpan.className = "tab-index-close-disabled";
					if(this.closeTabEvent)
						oCloseSpan.removeEventListener("click", this.closeTabEvent, false);
				}
			}
			this.selectTabEvent = null;
			this.closeTabEvent = null;
			
			if(jaswin.button)
			{
				if(this.indexMouseMoveEvent)
					this.index.removeEventListener("mouseover", this.indexMouseMoveEvent, false);
				if(this.indexMouseOutEvent)
					this.index.removeEventListener("mouseout", this.indexMouseOutEvent, false);
			}
			this.indexMouseMoveEvent = null;
			this.indexMouseOutEvent = null;
		}
		
		//==================================================
		//	Add control to this container
		//==================================================
		addControl(oControl, oParentElement)
		{
			return;
		}
		
		//==================================================
		//	Remove control from this container
		//==================================================
		removeControl(oControl)
		{
			return;
		}
		
		//==================================================
		//	Get child panel
		//==================================================
		getChildPanel(iIndex)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"panel.js", "jaswin.panel.Tab.getChildPanel");
			if((iIndex < 0) || (iIndex >= this.indexes.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, this.id),
											"panel.js", "jaswin.panel.Tab.getChildPanel");
			
			return this.controls[iIndex];
		}
		
		//==================================================
		//	Set background color
		//==================================================
		setBackgroundColor(iIndex, sColor)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"panel.js", "jaswin.panel.Tab.setBgColor");
			if((iIndex < 0) || (iIndex >= this.indexes.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, this.id),
											"panel.js", "jaswin.panel.Tab.setBgColor");
			
			if(typeof sColor != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_BACKGROUND, this.id),
											"panel.js", "jaswin.panel.Tab.setBgColor");
			
			this.bgColors[iIndex] = sColor;
			if(iIndex == this.selectedIndex)
			{
				this.indexes[iIndex].style.backgroundColor = sColor;
				this.indexes[iIndex].style.borderBottomColor = sColor;
			}
			this.controls[iIndex].element.style.backgroundColor = sColor;
		}
		
		//==================================================
		//	Select a tab index
		//==================================================
		select(e)
		{
			let oElement = e.target;
			let iCounter = 0;
			let iIndex = -1;
			
			if(!oElement)
				return;
			
			for(iCounter = 0; iCounter < this.indexes.length; iCounter++)
			{
				if(this.indexes[iCounter].contains(oElement))
				{
					iIndex = iCounter;
					break;
				}
			}
			
			if(iIndex != -1)
				this.selectByIndex(iIndex);
		}
		
		//==================================================
		//	Select a tab by index
		//==================================================
		selectByIndex(iIndex)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"panel.js", "jaswin.panel.Tab.getChildPanel");
			if((iIndex < 0) || (iIndex >= this.indexes.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, this.id),
											"panel.js", "jaswin.panel.Tab.getChildPanel");
			
			let iCounter = 0;
			let iOrgIndex = this.selectedIndex;
			let bFound = false;
			let iLeft = 0;
			let iRight = 0;
			let iScroll = 0;
			
			for(iCounter = 0; iCounter < this.indexes.length; iCounter++)
			{
				if(iCounter == iIndex)
				{
					this.indexes[iCounter].className = "tab-index-on";
					if(this.bgColors[iCounter] != "")
					{
						this.indexes[iCounter].style.backgroundColor = this.bgColors[iCounter];
						this.indexes[iCounter].style.borderBottom = "solid 1px " + this.bgColors[iCounter];
					}
					if(this.adjustIndexHeight && (iCounter != iOrgIndex))
						this.indexes[iCounter].style.paddingTop = parseInt(this.indexes[iCounter].style.paddingTop) + 2 + "px";
					
					this.controls[iCounter].element.style.display = "block";
					this.selectedIndex = iCounter;
					bFound = true;
				}
				else
				{
					this.indexes[iCounter].className = "tab-index-off";
					
					if(this.adjustIndexHeight && (iCounter == iOrgIndex))
						this.indexes[iCounter].style.paddingTop = parseInt(this.indexes[iCounter].style.paddingTop) - 2 + "px";
					
					if(this.bgColors[iCounter] != "")
						this.indexes[iCounter].style.borderBottom = "none";
					
					this.controls[iCounter].element.style.display = "none";
					if(!bFound)
						iLeft += this.indexes[iCounter].offsetWidth;
				}
			}
			
			iRight = iLeft + this.indexes[iIndex].offsetWidth;
			if(iLeft - this.indexContainer.scrollLeft < 0)
			{
				iScroll = iLeft;
				this.indexContainer.scrollLeft = iScroll;
			}
			else if(iRight - this.indexContainer.scrollLeft > this.index.offsetWidth)
			{
				iScroll = iRight - this.index.offsetWidth + 2;
				this.indexContainer.scrollLeft = iScroll;
			}
			
			if(this.onSelect)
				this.onSelect(this.selectedIndex);
		}
		
		//==================================================
		//	Find tab index with label
		//==================================================
		findIndex(sLabel)
		{
			let iCounter = 0;
			for(iCounter = 0; iCounter < this.indexes.length; iCounter++)
			{
				if(this.indexes[iCounter].childNodes[0].nodeValue == sLabel)
					return iCounter;
			}
			
			return -1;
		}
		
		//==================================================
		//	Mouse move a tab index
		//==================================================
		hoverIndex()
		{
			let iLastScroll = 0;
			let iCurScroll = this.indexContainer.scrollLeft;
			let iIndexWidth = this.index.offsetWidth;
			let iInnerWidth = 0;
			let iCounter;
			
			for(iCounter = 0; iCounter < this.indexes.length; iCounter++)
				iInnerWidth += this.indexes[iCounter].offsetWidth;
			iLastScroll = iInnerWidth - iIndexWidth + 2;
			
			if(iCurScroll)
				this.scrollLeftButton.show();
			if(iInnerWidth > iIndexWidth)
			{
				if(iCurScroll != iLastScroll)
					this.scrollRightButton.show();
			}
		}
		
		//==================================================
		//	Mouse out a tab index
		//==================================================
		outIndex(e)
		{
			let oElement = e.target || e.srcElement;
			let oRect = this.index.getBoundingClientRect();
			
			if((e.clientX < oRect.left)
			|| (e.clientX > oRect.right)
			|| (e.clientY < oRect.top)
			|| (e.clientY > oRect.bottom)
			|| (this.indexInner.contains(oElement)))
			{
				this.scrollLeftButton.hide();
				this.scrollRightButton.hide();
			}
		}
		
		//==================================================
		//	Scroll left a tab index
		//==================================================
		scrollLeft()
		{
			let iCurScroll = this.indexContainer.scrollLeft;
			
			if(iCurScroll - 64 < 0)
				this.indexContainer.scrollLeft = 0;
			else
				this.indexContainer.scrollLeft = iCurScroll - 64;
		}
		
		//==================================================
		//	Scroll right a tab index
		//==================================================
		scrollRight()
		{
			let iLastScroll = 0;
			let iCurScroll = this.indexContainer.scrollLeft;
			let iIndexWidth = this.index.offsetWidth;
			let iInnerWidth = 0;
			let iCounter;
			
			for(iCounter = 0; iCounter < this.indexes.length; iCounter++)
				iInnerWidth += this.indexes[iCounter].offsetWidth;
			iLastScroll = iInnerWidth - iIndexWidth + 2;
			
			if(iCurScroll + 64 > iLastScroll)
				this.indexContainer.scrollLeft = iLastScroll;
			else
				this.indexContainer.scrollLeft += 64;
		}
		
		//==================================================
		//	Add tab
		//==================================================
		addTab(sLabel)
		{
			// Check parameter
			if(!this.style & panel.STYLE_VARIABLE)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, this.id),
											"panel.js", "jaswin.panel.Tab.addTab");
			
			if(typeof sLabel != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, this.id),
											"panel.js", "jaswin.panel.Tab.addTab");
			
			// Get ID
			let iTmp = 0;
			let sID = null;
			let iCounter = 0;
			let bFound;
			
			while(true)
			{
				sID = this.id + "_" + iTmp;
				bFound = false;
				for(iCounter = 0; iCounter < this.controls.length; iCounter++)
				{
					if(this.controls[iCounter].id == sID)
					{
						bFound = true;
						break;
					}
				}
				if(!bFound)
					break;
				
				iTmp++;
			}
			
			// Create content
			let oExpRect = new jaswin.ExpandableRect(jaswin.ANCHOR_BOTH, jaswin.ANCHOR_BOTH,
													0, 0,
													0, 0,
													0, 0);
			let oPanel = new panel.Panel(sID, oExpRect,
										panel.STYLE_DEFAULT, true,
										panel.OVERFLOW_DEFAULT);
			if(this.indexes.length == 0)
				this.selectedIndex = 0;
			else
				oPanel.hide();
			this.container.appendChild(oPanel.element);
			this.controls.push(oPanel);
			oPanel.setParent(this);
			
			// Create index
			let oSpan = document.createElement("span");
			this.indexes.push(oSpan);
			if(this.indexes.length == 1)
			{
				if(this.enabled)
					oSpan.className = "tab-index-on";
				else
					oSpan.className = "tab-index-on-disabled";
			}
			else
			{
				if(this.enabled)
					oSpan.className = "tab-index-off";
				else
					oSpan.className = "tab-index-off-disabled";
			}
			oSpan.innerHTML = sLabel;
			
			let oCloseSpan = document.createElement("span");
			if(this.enabled)
				oCloseSpan.className = "tab-index-close";
			else
				oCloseSpan.className = "tab-index-close-disabled";
			oCloseSpan.innerHTML = "x";
			oSpan.appendChild(oCloseSpan);
			
			if(this.enabled)
			{
				if(this.closeTabEvent)
					oCloseSpan.addEventListener("click", this.closeTabEvent, false);
				else
				{
					oCloseSpan.addEventListener("click", this.closeTabEvent = (e) => {
						this.closeTab(e);
					}, false);
				}
				
				if(this.selectTabEvent)
					oSpan.addEventListener("click", this.selectTabEvent, false);
				else
				{
					oSpan.addEventListener("click", this.selectTabEvent = (e) => {
						this.select(e);
					}, false);
				}
			}
			
			this.indexInner.appendChild(oSpan);
			
			if(this.parent)
			{
				// Set index scroll button
				let iInnerWidth = 0;
				let iCounter;
				
				for(iCounter = 0; iCounter < this.indexes.length; iCounter++)
					iInnerWidth += this.indexes[iCounter].offsetWidth;
				
				this.indexInner.style.width = (iInnerWidth + 4) + "px";
			}
			
			this.bgColors.push("");
		}
		
		//==================================================
		//	Close tab
		//==================================================
		closeTab(e)
		{
			let oElement = e.target;
			let iCounter;
			let iIndex;
			
			// Find close tab
			for(iCounter = 0; iCounter < this.indexes.length; iCounter++)
			{
				if(this.indexes[iCounter].contains(oElement))
				{
					this.indexInner.removeChild(this.indexInner.childNodes[iCounter]);
					this.indexes.splice(iCounter, 1);
					
					this.controls[iCounter].destroy();
					this.controls.splice(iCounter, 1);
					this.bgColors.splice(iCounter, 1);
					
					iIndex = iCounter;
					break;
				}
			}
			
			// Set index scroll button
			let iInnerWidth = 0;
			
			for(iCounter = 0; iCounter < this.indexes.length; iCounter++)
				iInnerWidth += this.indexes[iCounter].offsetWidth;
			
			this.indexInner.style.width = (iInnerWidth + 4) + "px";
			
			// Reset selected tab
			if((iIndex == this.selectedIndex) && (this.indexes.length > 0))
			{
				if(iIndex >= this.indexes.length)
					iIndex--;
				
				this.indexes[iIndex].className = "tab-index-on";
				
				if(this.adjustIndexHeight)
					this.indexes[iIndex].style.paddingTop = parseInt(this.indexes[iIndex].style.paddingTop) + 2 + "px";
				
				this.controls[iIndex].element.style.display = "block";
				this.selectedIndex = iIndex;
			}
		}
	}
	
	
	//==========================================================
	//
	//	Function	:	tabIndexScrollLeft
	//
	//	Summary		:	Event handler of scroll left a index
	//
	//==========================================================
	function tabIndexScrollLeft(oBtn)
	{
		oBtn.parent.scrollLeft();
	}
	
	
	//==========================================================
	//
	//	Function	:	tabIndexScrollRight
	//
	//	Summary		:	Event handler of scroll right a index
	//
	//==========================================================
	function tabIndexScrollRight(oBtn)
	{
		oBtn.parent.scrollRight();
	}
	
}(jaswin.panel = jaswin.panel || {}));
